# Copyright (C) 2014 Red Hat
#
# This file is part of relval.
#
# wikitcms is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author: Adam Williamson <awilliam@redhat.com>

# relval interacts with Fedora release validation events. This file contains
# functions that help implement the 'testcase-stats' sub-command of relval,
# which produces statistics about release validation test coverage in HTML
# format.

from collections import defaultdict, Counter
from operator import attrgetter
import re
import os
import sys
import datetime
import shutil
from pkg_resources import resource_string, resource_filename

SANITY_SUB = re.compile('\W')

class Test(object):
    """An object representing information for one 'unique test' (see
    main() for definition). Counts total passes, fails and warns for
    the test across all composes as 'p', 'f' and 'w', and creates a
    'bitmap' of these totals as a dict keyed by compose, each entry
    being a list with two entries, the first being the total of results
    and the second a string that is used to style the HTML output
    depending on if the results are all passes, all fails or a mix.
    Keeps a note of the most recent compose for which it was sent a
    result row: this relies on the result rows being sent in the
    correct order, this class does not know how to order the results it
    is sent.

    Stores lists of the wikitcms result objects for this test in
    three more dicts, one each for 'passes', 'fails' and warns', keyed
    by compose. As it extracts result objects from the result row, it
    notes the environments for which any results were provided, and
    divides this into the number of 'boxes' (i.e. the total number of
    environments for which results could have been provided) to produce
    a coverage percentage for each compose. This information is used to
    generate the detailed result pages. 
    """
    def __init__(self):
        self.passes = defaultdict(list)
        self.warns = defaultdict(list)
        self.fails = defaultdict(list)
        self.coverage = defaultdict(int)
        self.composes = set()
        self.bitmap = dict()
        self.last_tested = ""
        self.p = 0
        self.f = 0
        self.w = 0

    def update(self, compose, row):
        """Passed a result row and the 2-tuple that identifies the compose
        it comes from. Parses the data from the result row as described in
        the class description. Assumes there will only ever be one result
        row per unique test per compose: if this is ever not true, the
        bitmap will represent only the results from the last row seen, and
        the coverage percentage will be inaccurate.
        """
        self.milestone = row.milestone
        self.composes.add(compose)
        rlists = row.results.iteritems()
        p = self.p
        f = self.f
        w = self.w
        envs = 0
        for env, rlist in rlists:
            if rlist:
                envs += 1
            for result in rlist:
                if result.status == 'pass':
                    self.last_tested = compose
                    self.passes[compose].append((env, result))
                    p += 1
                if result.status == 'fail':
                    self.last_tested = compose
                    self.fails[compose].append((env, result))
                    f += 1
                if result.status == 'warn':
                    self.last_tested = compose
                    self.warns[compose].append((env, result))
                    w += 1

        self.coverage[compose] = envs*100 // row.boxes
        self.bitmap[compose] = [0, 'nottested']
        if p+f+w > 0:
            b = self.bitmap[compose]
            if f+w == 0: # just passed
                b[1] = 'pass'
            elif p+w == 0: # just failed
                b[1] = 'fail'
            elif w > 0:
                b[1] = 'warn'
            b[0] += p+f+w

    def get_page_contents(self, compose):
        """Generates the content for the detailed result pages for each
        'unique test', which are linked from the summary page.
        """
        body = ''
        if compose not in self.composes:
            # There's a page for this compose and test type, but this
            # 'unique test' doesn't seem to exist in it (happens if
            # test got renamed or moved sections).
            body += "<b>Test not found for compose.</b>\n</p>\n"
            return body
        if self.coverage[compose] == 0:
            body += "<b>No results.</b>\n</p>\n"
            return body
        body += "<b>Coverage:</b> {}%<br/>".format(self.coverage[compose])
        # 'g' lets us get the results for the status we're currently
        # working on. results for each status are stored as tuples of the
        # result environment and the result object itself.
        #
        #'c' counts the occurrences of each environment in the set of
        # results for the status; the following line reads out the list of
        # envs 'c' found and prints the name of each env and the count of
        # results that was found for it.
        for status in ('Passes', 'Warns', 'Fails'):
            body += "<b>{}: </b>".format(status)
            g = attrgetter(status.lower())
            c = Counter([env for (env, res) in g(self)[compose]])
            body += ', '.join(["{} - <b>{}</b>".format(env, c[env]) for
                              env in c])
            body += "\n<ul>"
            for (env, result) in g(self)[compose]:
                try:
                # Add a list of bugs for this status, if any.
                    url = ('{0}: <a href="https://bugzilla.redhat.com'
                           'bugzilla/show_bug.cgi?id={1}">{1}</a>')
                    body += "\n<li>" + '\n<li>'.join([url.format(
                            result.user, bug) for bug in result.bugs])
                except (IndexError, TypeError):
                    pass
            body += "\n</ul>\n"
        return body


def get_bitmap_html(bitmap, allcomposes):
    """Renders the HTML for the little graphic that illustrates test
    coverage for a given 'unique test' in the final output. Uses the
    data from the Test() object's bitmap property, which is a dict of
    tuples keyed by compose, each tuple providing a result count and
    a string indicating whether any results are all passes, all fails,
    or a mix of the two; that string as a CSS class makes the dot(s)
    green, red or orange (or grey for 'nottested' if 0 results). If
    we believe the result row is entirely missing for a compose, an
    empty space will be rendered ('sym_nopage').
    """
    out = '<span style="letter-spacing:-0.5em;font-size:1.5em;">'
    # These are the little 'dot' symbols the coverage bitmap uses.
    sym_nopage = "&#10240;"
    sym_0 = "&#10244;"
    sym_1 = "&#10244;"
    sym_2 = "&#10246;"
    sym_3 = "&#10247;"
    for compose in allcomposes:
        try:
            (cnt, style) = bitmap[compose]
            out += '<span class="{}">'.format(style)
            if cnt == 0:
                out+= sym_0
            elif cnt < 3:
                out+= sym_1
            elif cnt < 5:
                out+= sym_2
            else:
                out+= sym_3
            out+='</span>'
        except:
            # if that failed, it means this test did not exist for this compose.
            out += sym_nopage
    out += '</span>'
    return out

def print_results_html(testtype, pages, allcomposes, tests, outdir):
    """Produces the summary page for each 'testtype' (Installation,
    Base etc).
    """
    # directory with templates for pages
    #htmldir = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
    #                       'html')

    # template-y bits used to construct the page HTML
    col_names = ['Testcase URL','Release Level', 'Last in', 'Coverage',
                 'Details']
    trs = []
    thead =  ('<h4>{0}</h4>\n\t<table id="testcases-table-{0}"'
              'class="tablesorter" border="0" cellpadding="0" cellspacing="1">'
              '\n\t\t<thead>\n\t\t\t<tr>'+"\n".join(
                    ["<th>{}</th>".format(c) for c in col_names])+
              '\n\t\t\t</tr>\n\t\t</thead>\n\t\t<tbody>')
    tend = '\t\t</tbody>\n\t</table>'
    row_pre = ('<td style="padding=0"><a href="https://fedoraproject.org/'
              'wiki/{testcase}">{testcase}{name}</a></td>')
    row_main = '<td style="padding=0">{}</td>\n'* (len(col_names)-2)
    row_post = '<td style="padding=0"><a href="{link}.html">Details</a></td>\n'

    lastsec = ''
    firsttest = True

    # this is where we actually start doin' stuff
    os.makedirs(os.path.join(outdir, testtype))
    for (name, testcase, section), test in tests.iteritems():
        # If this is the very first row in the page, create the first table
        # header.
        if firsttest:
            trs.append(thead.format(section))
            firsttest = False
        # Otherwise, start a new table if we've changed section.
        elif section != lastsec:
            trs.append(tend)
            trs.append(thead.format(section))
        # Update the variable that records the last section we saw.
        lastsec = section
        # We use slightly different formats for both the detail page filenames
        # and the summary page visible names for tests whose 'name' differs
        # from their 'testcase'. We want to display and use both attributes
        # when they differ, but just use one when they're the same.
        if testcase != name:
            dispname = ' - ' + name
            linkname = '-' + name
        else:
            dispname = ''
            linkname = ''
        # If there's a section for the test case, include it in the page URL
        # in a nice way.
        if section:
            secturl = '-' + section
        else:
            secturl = ''
        tr = row_pre.format(testcase=testcase, name=dispname)
        tr += row_main.format(test.milestone, ' '.join(test.last_tested),
                              get_bitmap_html(test.bitmap, allcomposes))
        tr += row_post.format(link=testtype+'/'+SANITY_SUB.sub('_', testcase)+
                              SANITY_SUB.sub('_', linkname)+
                              SANITY_SUB.sub('_', secturl))
        trs.append("<tr>" + tr + "</tr>")

        # write detailed result page for this test
        template = resource_string(__name__, 'html/testcase.html')
        body = ""
        for page in pages:
            body += ('<p><strong><a href="https://fedoraproject.org/wiki/{0}">'
                     '{0}</a></strong><br />\n').format(page.wikiname)
            body += test.get_page_contents((page.milestone, page.compose))
            body += '</p>'
        fname = os.path.join(outdir, testtype, SANITY_SUB.sub('_', testcase)+
                         SANITY_SUB.sub('_', linkname)+
                         SANITY_SUB.sub('_', secturl)+'.html')
        try:
            fout = open(fname, 'w')
            if testcase != name:
                outname = ' - ' + name
            else:
                outname = ''
            fout.write(template.format(timestamp=datetime.datetime.utcnow(),
                       testcase=testcase, name=outname, body=body))
            fout.close()
        except:
            pass

    trs.append(tend)
    body = "\n".join(trs)
    # create the Overview page
    print("Creating HTML output: " + os.path.join(outdir, testtype + '.html'))
    template = resource_string(__name__, 'html/index.html')
    fout = open(os.path.join(outdir, testtype + '.html'), 'w')
    fout.write(template % {'timestamp': datetime.datetime.utcnow(), 'body': body})
    fout.close()

def print_summary_html(outdir, allpages):
    # copy javascript
    htmldir = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), 'html')
    resource_filename(__name__, 'html/js')
    shutil.copytree(resource_filename(__name__, 'html/js'), os.path.join(outdir, 'js'))

    # Create index page
    fout = open(os.path.join(outdir, "index.html"), "w")
    fout.write("<html><body>\n")
    fout.write("<p>Created at: %s (UTC)</p>" % datetime.datetime.utcnow())
    for testtype in allpages.iterkeys():
        fname = testtype + '.html'
        fout.write('<a href="{0}.html">{0}</a><br />'.format(testtype))
    fout.write("\n</body></html>")
    fout.close()
